@@ -1,25 +1,13 @@
+import os
 import sys
 
-from pathlib import PureWindowsPath as PurePath, WindowsPath as Path
-from ._types import PydFile, File, LiteralXML, Property, ItemDefinition, ConditionalValue
+from pathlib import PurePath, Path
+from ._types import Package, PydFile, File, LiteralXML, Property, ItemDefinition, ConditionalValue
 from ._writer import ProjectFileWriter
 
 LIBPATH = Path(sys.base_prefix) / "libs"
 INCPATH = Path(sys.base_prefix) / "include"
-
-
-def _all_members(item, recurse_if=None, return_if=None, *, prefix=""):
-    if not return_if or return_if(item):
-        yield "{}{}".format(prefix, item.name), item
-    if not recurse_if or recurse_if(item):
-        for m in item.members:
-            yield from _all_members(
-                m,
-                recurse_if,
-                return_if,
-                prefix="{}{}/".format(prefix, item.name),
-            )
-
+SEP = os.path.sep
 
 class GroupSwitcher:
     def __init__(self, project):
@@ -46,36 +34,71 @@ class GroupSwitcher:
             return self._cm.__exit__(*exc_info)
 
 
-def _resolve_wildcards(basename, source):
+def _resolve_wildcards(basename, source, include_dirs=True):
     if source.parent.name == "**":
-        for d in source.parent.parent.iterdir():
-            if d.is_dir():
-                yield from _resolve_wildcards(basename, d / source.name)
+        name = PurePath(basename).parent
+        root = source.parent.parent
+        for p in root.iterdir():
+            if p.is_dir():
+                yield from _resolve_wildcards(
+                    name / p.relative_to(root) / source.name,
+                    p / "**" / source.name,
+                    include_dirs=False,
+                )
+            else:
+                yield name / p.name, p
+    elif "**" in source.parts:
+        raise ValueError("Unsupported wildcard pattern" + str(source))
     elif "*" in source.name or "?" in source.name:
-        name = PurePath(basename)
+        name = PurePath(basename).parent
         for p in source.parent.glob(source.name):
-            yield name.parent / p.name, p
-    else:
+            if not p.is_dir():
+                yield name / p.name, p
+    elif any("*" in p or "?" in p for p in source.parts):
+        raise ValueError("Unsupported wildcard pattern " + str(source))
+    elif include_dirs or not source.is_dir():
         yield basename, source
 
+
+def _all_members(item, recurse_if=None, return_if=None, *, prefix=""):
+    if not return_if or return_if(item):
+        yield "{}{}".format(prefix, item.name), item
+    if not recurse_if or recurse_if(item):
+        for m in item.members:
+            yield from _all_members(
+                m,
+                recurse_if,
+                return_if,
+                prefix="{}{}/".format(prefix, item.name),
+            )
+
+
 def _write_members(f, source_dir, members):
     with GroupSwitcher(f) as g:
         for n, p in members:
             if isinstance(p, File):
                 g.switch_to("ItemGroup")
                 wrote_any = False
+                exclude = ()
+                condition = None
+                if getattr(p, "has_condition", False):
+                    condition = getattr(p, "condition", None)
+                    pattern = getattr(p, "exclude", None)
+                    if pattern:
+                        exclude = {p2 for n2, p2 in _resolve_wildcards(n, source_dir / pattern)}
                 for n2, p2 in _resolve_wildcards(n, source_dir / p.source):
+                    if p2 in exclude:
+                        continue
                     options = dict(p.options)
-                    try:
-                        options.setdefault("RelativeSource", p2.relative_to(source_dir))
-                    except ValueError:
-                        pass
                     options.setdefault("SourceDir", source_dir)
                     options.setdefault("Name", n2)
+                    if condition:
+                        p2 = ConditionalValue(p2, condition=condition)
                     f.add_item(p._ITEMNAME, p2, **options)
                     wrote_any = True
                 if not wrote_any:
-                    raise ValueError("failed to find any files for " + str(p.source))
+                    raise ValueError("failed to find any files for {} in {}".format(
+                        p.source, source_dir))
             elif isinstance(p, Property):
                 g.switch_to("PropertyGroup")
                 f.add_property(p.name, p.value)
@@ -87,6 +110,8 @@ def _write_members(f, source_dir, members):
             elif isinstance(p, LiteralXML):
                 g.switch_to(None)
                 f.add_text(p.xml)
+            elif hasattr(p, "write_member"):
+                p.write_member(f, g)
 
 
 def _generate_pyd(project, build_dir, root_dir):
@@ -98,61 +123,84 @@ def _generate_pyd(project, build_dir, root_dir):
     if project.project_file:
         return Path(project.project_file)
 
-    with ProjectFileWriter(proj, project.name, vc_platforms=True) as f:
+    tpath = project.options.get("TargetName", project.name) + (project.options.get("TargetExt") or ".pyd")
+    tname, tdot, text = tpath.rpartition(".")
+    with ProjectFileWriter(proj, tname, vc_platforms=True, root_namespace=project.name) as f:
         with f.group("PropertyGroup", Label="Globals"):
             f.add_property("SourceDir", ConditionalValue(source_dir, if_empty=True))
             f.add_property("SourceRootDir", ConditionalValue(root_dir, if_empty=True))
-            f.add_property("OutDir", "layout\\")
-            f.add_property("IntDir", ConditionalValue("build\\", if_empty=True))
+            f.add_property("OutDir", f"layout{SEP}")
+            f.add_property("IntDir", ConditionalValue(f"build{SEP}", if_empty=True))
+            f.add_property("__TargetExt", tdot + text)
             for k, v in project.options.items():
                 if k not in {"ConfigurationType", "TargetExt"}:
                     f.add_property(k, v)
-        f.add_import(r"$(VCTargetsPath)\Microsoft.Cpp.Default.props")
-        f.add_import(r"$(PyMsbuildTargets)\common.props")
+        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-default-$(Platform).props")
+        f.add_import(f"$(PyMsbuildTargets){SEP}common.props")
         with f.group("PropertyGroup", Label="Configuration"):
             f.add_property("ConfigurationType", project.options.get("ConfigurationType", "DynamicLibrary"))
             f.add_property("PlatformToolset", "$(DefaultPlatformToolset)")
+            f.add_property("BasePlatformToolset", "$(DefaultPlatformToolset)")
             f.add_property("CharacterSet", "Unicode")
-        f.add_import(r"$(VCTargetsPath)\Microsoft.Cpp.props")
-        with f.group("PropertyGroup"):
-            f.add_property("TargetExt", project.options.get("TargetExt", ".pyd"))
-            f.add_property("LinkIncremental", "false")
-        with f.group("ItemDefinitionGroup"):
-            with f.group("ClCompile"):
-                f.add_property("AdditionalIncludeDirectories", "{};%(AdditionalIncludeDirectories)".format(INCPATH)),
-            with f.group("Link"):
-                f.add_property("AdditionalLibraryDirectories", "{};%(AdditionalLibraryDirectories)".format(LIBPATH)),
-                f.add_property("GenerateDebugInformation", "True")
+        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-$(Platform).props")
+        f.add_import(f"$(PyMsbuildTargets){SEP}pyd.props")
 
         _write_members(f, source_dir, _all_members(project, recurse_if=lambda m: m is project))
+        for n, p in _all_members(project, recurse_if=lambda m: m is project, return_if=lambda m: isinstance(m, Package)):
+            _write_members(
+                f,
+                source_dir,
+                _all_members(p, recurse_if=lambda m: not isinstance(m, PydFile), prefix=f"{project.name}/")
+            )
 
-        f.add_import(r"$(PyMsbuildTargets)\common.targets")
-        f.add_import(r"$(VCTargetsPath)\Microsoft.Cpp.targets")
-        f.add_import(r"$(PyMsbuildTargets)\pyd.targets")
+        f.add_import(f"$(PyMsbuildTargets){SEP}common.targets")
+        f.add_import(f"$(PyMsbuildTargets){SEP}cpp-$(Platform).targets")
+        f.add_import(f"$(PyMsbuildTargets){SEP}pyd.targets")
 
     return proj
 
 
-def generate(project, build_dir, source_dir):
+def _generate_pyd_reference_metadata(relname, project, source_dir):
+    output = project.options.get("TargetName", relname.stem) + project.options.get("TargetExt", ".pyd")
+    tname, tdot, text = output.rpartition(".")
+    return {
+        **dict(
+            TargetDir=relname.parent,
+            IntDir="$(IntDir)" + tname,
+        ),
+        **project.options,
+        **dict(
+            TargetName=tname,
+            TargetExt=tdot + text,
+            SourceDir=source_dir / project.source,
+        ),
+    }
+
+
+def generate(project, build_dir, source_dir, config_file=None):
     build_dir = Path(build_dir)
     root_dir = Path(source_dir)
+    config_file = root_dir / (config_file or "_msbuild.py")
     source_dir = root_dir / project.source
     proj = build_dir / "{}.proj".format(project.name)
 
     if project.project_file:
         return Path(project.project_file)
 
+    if isinstance(project, PydFile):
+        return _generate_pyd(project, build_dir, root_dir)
+
     with ProjectFileWriter(proj, project.name) as f:
         with f.group("PropertyGroup"):
             f.add_property("SourceDir", ConditionalValue(source_dir, if_empty=True))
             f.add_property("SourceRootDir", ConditionalValue(root_dir, if_empty=True))
-            f.add_property("OutDir", ConditionalValue("layout\\", if_empty=True))
-            f.add_property("IntDir", ConditionalValue("build\\", if_empty=True))
+            f.add_property("OutDir", ConditionalValue(f"layout{SEP}", if_empty=True))
+            f.add_property("IntDir", ConditionalValue(f"build{SEP}", if_empty=True))
             for k, v in project.options.items():
                 f.add_property(k, v)
-        f.add_import(r"$(PyMsbuildTargets)\common.props")
+        f.add_import(f"$(PyMsbuildTargets){SEP}common.props")
         with f.group("ItemGroup", Label="ProjectReferences"):
-            for n, p in _all_members(project, return_if=lambda m: isinstance(m, PydFile)):
+            for n, p in _all_members(project, return_if=lambda m: m is not project and isinstance(m, PydFile)):
                 fn = PurePath(n)
                 pdir = _generate_pyd(p, build_dir, source_dir)
                 try:
@@ -163,26 +211,18 @@ def generate(project, build_dir, source_dir):
                     "Project",
                     pdir,
                     Name=n,
-                    **{
-                        **dict(
-                            TargetDir=fn.parent,
-                            TargetName=fn.stem,
-                            TargetExt=".pyd",
-                            SourceDir=source_dir / p.source,
-                        ),
-                        **p.options,
-                    }
+                    **_generate_pyd_reference_metadata(fn, p, source_dir),
                 )
         with f.group("ItemGroup", Label="Sdist metadata"):
             f.add_item("Sdist", build_dir / "PKG-INFO", RelativeSource="PKG-INFO")
-            f.add_item("Sdist", root_dir / "_msbuild.py", RelativeSource="_msbuild.py")
+            f.add_item("Sdist", config_file, RelativeSource="_msbuild.py")
             f.add_item("Sdist", root_dir / "pyproject.toml", RelativeSource="pyproject.toml")
         _write_members(f, source_dir, _all_members(
             project,
             recurse_if=lambda m: not isinstance(m, PydFile),
         ))
-        f.add_import(r"$(PyMsbuildTargets)\common.targets")
-        f.add_import(r"$(PyMsbuildTargets)\package.targets")
+        f.add_import(f"$(PyMsbuildTargets){SEP}common.targets")
+        f.add_import(f"$(PyMsbuildTargets){SEP}package.targets")
 
     return proj
 
