@@ -1,5 +1,5 @@
-from os import name
-from pathlib import WindowsPath as Path, PureWindowsPath as PurePath
+from importlib.machinery import EXTENSION_SUFFIXES as _EXTENSION_SUFFIXES
+from pathlib import Path, PurePath
 
 __all__ = [
     "Package",
@@ -14,12 +14,12 @@ __all__ = [
     "CSourceFile",
     "IncludeFile",
     "File",
+    "RemoveFile",
 ]
 
 
 class _Project:
     r"""Base class of compilable projects. Do not use directly."""
-    _ACTIONS = ()
     options = {}
 
     def __init__(self, name, *members, project_file=None, source="", **options):
@@ -29,6 +29,16 @@ class _Project:
         self.project_file = project_file
         self.members = list(members)
 
+    def __iter__(self):
+        for m in self.members:
+            yield m
+            try:
+                it = iter(m)
+            except TypeError:
+                pass
+            else:
+                yield from it
+
 
 class Package(_Project):
     r"""Represents a Python package.
@@ -69,7 +79,7 @@ Specify `source` to find sources in a subdirectory.
 
 Other options will be added to the project as properties.
 """
-    options = {"TargetExt": ".pyd"}
+    options = {"TargetExt": None}
 
 
 class LiteralXML:
@@ -95,7 +105,8 @@ Specify `condition` to directly control the test, or `if_empty` to
 check for an existing value.
 
 Specify either `prepend` or `append` (or both, if you have a reason)
-to concatenate the existing property value.
+to concatenate the existing property value. Remember to include the
+appropriate separator character in your value.
 """
     has_condition = True
 
@@ -119,6 +130,15 @@ to concatenate the existing property value.
         return str(self.value)
 
 
+def Prepend(value):
+    r"""Add a value that will prepend any existing value.
+
+This is a helper function for ConditionalValue that sets `prepend` to True.
+Remember to include the appropriate separator character in your value.
+"""
+    return ConditionalValue(value, prepend=True)
+
+
 class Property:
     r"""Add a property to a project.
 
@@ -164,6 +184,8 @@ project treats "Content" elements.
 """
     _ITEMNAME = "Content"
     options = {}
+    has_condition = False
+    condition = None
 
     def __init__(self, source, name=None, **metadata):
         self.source = PurePath(source)
@@ -171,6 +193,16 @@ project treats "Content" elements.
         self.members = []
         self.options = {**self.options, **metadata}
 
+    def excluding(self, pattern):
+        self.has_condition = True
+        self.exclude = pattern
+        return self
+
+    def if_(self, condition):
+        self.has_condition = True
+        self.condition = condition
+        return self
+
 
 class PyFile(File):
     r"""Add a Python source file to a package.
@@ -207,11 +239,23 @@ class CSourceFile(File):
 These files will be included in the sdist, but will not be copied
 in-place or included in wheels, except as built output.
 
-Incremental rebuilds will be triggered when these files are modified.
+Incremental rebuilds will be triggered when these files are modified,
+and each file is expected to produce a linkable file.
 """
     _ITEMNAME = "ClCompile"
 
 
+class LinkFile(File):
+    r"""Add a linker input file.
+
+These files will be included in the sdist, but will not be copied
+in-place or included in wheels, except as built output.
+
+Incremental rebuilds will be triggered when these files are modified,
+and each file will be linked into the final output.
+"""
+    _ITEMNAME = "Link"
+
 class IncludeFile(File):
     r"""Add a header file.
 
@@ -219,7 +263,29 @@ These files will be included in the sdist, but will not be copied
 in-place or included in wheels, except as built output.
 
 Incremental rebuilds will be triggered when these files are modified,
-but they do not directly participate in most builds.
+but they do not produce linkable outputs.
 """
     _ITEMNAME = "ClInclude"
 
+class RemoveFile:
+    r"""Removes a file that has already been added.
+
+This must appear after the files that were added, and must specify the
+same kind (either the class used to create it, or the string literal
+matching the MSBuild item name).
+"""
+    _ITEMNAME = None
+    name = None
+    members = ()
+    has_condition = True
+    condition = None
+
+    def __init__(self, kind, pattern):
+        self._ITEMNAME = getattr(kind, "_ITEMNAME", None) or str(kind)
+        self.exclude = pattern
+
+    def if_(self, condition):
+        self.condition = condition
+
+    def __str__(self):
+        return ""
